"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.typedJsonCursor = exports.cursorToJsonObject = exports.singleJsonCursor = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../../core/index.js");
const index_js_2 = require("../../feature-libraries/index.js");
const index_js_3 = require("../../util/index.js");
const leafDomain_js_1 = require("../leafDomain.js");
const jsonDomainSchema_js_1 = require("./jsonDomainSchema.js");
const internal_2 = require("@fluidframework/runtime-utils/internal");
const adapter = {
    value: (node) => node !== null && typeof node === "object"
        ? undefined // arrays and objects have no defined value
        : node, // null, boolean, numbers, and strings are their own values
    type: (node) => {
        const type = typeof node;
        switch (type) {
            case "number":
                return leafDomain_js_1.leaf.number.name;
            case "string":
                return leafDomain_js_1.leaf.string.name;
            case "boolean":
                return leafDomain_js_1.leaf.boolean.name;
            default:
                if (node === null) {
                    return leafDomain_js_1.leaf.null.name;
                }
                else if ((0, index_js_3.isReadonlyArray)(node)) {
                    return jsonDomainSchema_js_1.jsonArray.name;
                }
                else {
                    return jsonDomainSchema_js_1.jsonObject.name;
                }
        }
    },
    keysFromNode: (node) => {
        switch (typeof node) {
            case "object":
                if (node === null) {
                    return [];
                }
                else if ((0, index_js_3.isReadonlyArray)(node)) {
                    return node.length === 0 ? [] : [index_js_1.EmptyKey];
                }
                else {
                    return Object.keys(node);
                }
            default:
                return [];
        }
    },
    getFieldFromNode: (node, key) => {
        // Object.prototype.hasOwnProperty can return true for strings (ex: with key "0"), so we have to filter them out.
        // Rather than just special casing strings, we can handle them with an early return for all primitives.
        if (typeof node !== "object") {
            return [];
        }
        if (node === null) {
            return [];
        }
        if ((0, index_js_3.isReadonlyArray)(node)) {
            return key === index_js_1.EmptyKey ? node : [];
        }
        if (Object.prototype.hasOwnProperty.call(node, key)) {
            const field = node[key];
            (0, internal_1.assert)(field !== undefined, 0x41e /* explicit undefined fields should not be preserved in JSON */);
            return [field];
        }
        return [];
    },
};
/**
 * Used to read generic json compatible data as a tree in the JSON domain.
 * The returned tree will have a schema in the json domain as defined by {@link jsonRoot}.
 *
 * @returns an {@link ITreeCursorSynchronous} for a single {@link JsonCompatible}.
 */
function singleJsonCursor(root) {
    return (0, index_js_2.stackTreeNodeCursor)(adapter, root);
}
exports.singleJsonCursor = singleJsonCursor;
/**
 * Extract a JS object tree from the contents of the given ITreeCursor.
 * Assumes that ITreeCursor contains only unaugmented JsonTypes.
 */
function cursorToJsonObject(reader) {
    const type = reader.type;
    switch (type) {
        case leafDomain_js_1.leaf.number.name:
        case leafDomain_js_1.leaf.boolean.name:
        case leafDomain_js_1.leaf.string.name:
            (0, internal_1.assert)(reader.value !== undefined, 0x84f /* out of schema: missing value */);
            (0, internal_1.assert)(!(0, internal_2.isFluidHandle)(reader.value), 0x850 /* out of schema: unexpected FluidHandle */);
            return reader.value;
        case jsonDomainSchema_js_1.jsonArray.name: {
            reader.enterField(index_js_1.EmptyKey);
            const result = (0, index_js_1.mapCursorField)(reader, cursorToJsonObject);
            reader.exitField();
            return result;
        }
        case jsonDomainSchema_js_1.jsonObject.name: {
            const result = {};
            (0, index_js_1.mapCursorFields)(reader, (cursor) => {
                const key = cursor.getFieldKey();
                (0, internal_1.assert)(cursor.firstNode(), 0x420 /* expected non-empty field */);
                // like `result[key] = cursorToJsonObject(reader);` except safe when keyString == "__proto__".
                Object.defineProperty(result, key, {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                    value: cursorToJsonObject(reader),
                });
                (0, internal_1.assert)(!cursor.nextNode(), 0x421 /* expected exactly one node */);
            });
            return result;
        }
        default: {
            (0, internal_1.assert)(type === leafDomain_js_1.leaf.null.name, 0x422 /* unexpected type */);
            return null;
        }
    }
}
exports.cursorToJsonObject = cursorToJsonObject;
const typedJsonSymbol = Symbol("JSON Cursor Type");
function isObject(json) {
    if (typeof json === "object" && json !== null) {
        const typed = json;
        return typed[typedJsonSymbol] !== undefined;
    }
    return false;
}
const typedAdapter = {
    value: (node) => {
        if ((0, internal_2.isFluidHandle)(node)) {
            return node;
        }
        // TODO: Get rid of this cast by combining TypedJsonCompatible and JsonCompatible into a single cursor with flags to configure behavior.
        // This cast is necessary because TypedJsonCompatible can have Fluid handles underneath it, however, `adapter` only cares if `node` itself is a FluidHandle, which it is not.
        return adapter.value(node);
    },
    type: (node) => {
        if ((0, internal_2.isFluidHandle)(node)) {
            return leafDomain_js_1.leaf.handle.name;
        }
        if (isObject(node)) {
            const type = node[typedJsonSymbol];
            return type instanceof index_js_2.TreeNodeSchemaBase ? type.name : (0, index_js_3.brand)(type);
        }
        return adapter.type(node);
    },
    keysFromNode: (node) => {
        if ((0, internal_2.isFluidHandle)(node)) {
            return [];
        }
        return adapter.keysFromNode(node);
    },
    getFieldFromNode: (node, key) => {
        if ((0, internal_2.isFluidHandle)(node)) {
            return [];
        }
        const field = adapter.getFieldFromNode(node, key);
        if ((0, index_js_3.isReadonlyArray)(field) && field.length === 1 && (0, index_js_3.isReadonlyArray)(field[0])) {
            // If the field is an array wrapping another array, then unbox to the inner array
            return field[0];
        }
        return field;
    },
};
/**
 * A variant of {@link singleJsonCursor} which allows types to be provided for nodes.
 *
 * @remarks Types are optional, but if present will be used to derive the type of the node when the cursor is read.
 *
 * This cursor differs from singleJsonCursor in that it inlines arrays, (arrays are not boxed into an "array node" but are directly interpreted as sequence fields).
 *
 * It also allows Fluid handles as input.
 *
 * @example
 * ```ts
 * const cursor = typedJsonCursor({
 *   [typedJsonCursor.type]: Point,
 *   x: 3,
 *   y: 42
 * });
 * ```
 */
const singleTypedJsonCursor = function (root) {
    return (0, index_js_2.stackTreeNodeCursor)(typedAdapter, root);
};
exports.typedJsonCursor = singleTypedJsonCursor;
singleTypedJsonCursor.type = typedJsonSymbol;
// #endregion TypedJsonCursor
//# sourceMappingURL=jsonCursor.js.map