"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
exports.unsupportedUsageError = exports.getOrCreateMapTreeNode = exports.tryGetMapTreeNode = exports.EagerMapTreeMapNode = exports.EagerMapTreeNode = exports.isMapTreeSequenceField = exports.isMapTreeNode = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../../core/index.js");
const index_js_2 = require("../../util/index.js");
const index_js_3 = require("../flex-tree/index.js");
const index_js_4 = require("../typed-schema/index.js");
const schemaBuilderBase_js_1 = require("../schemaBuilderBase.js");
const index_js_5 = require("../default-schema/index.js");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
/**
 * Checks if the given {@link FlexTreeNode} is a {@link MapTreeNode}.
 */
function isMapTreeNode(flexNode) {
    return flexNode instanceof EagerMapTreeNode;
}
exports.isMapTreeNode = isMapTreeNode;
/**
 * Checks if the given {@link FlexTreeField} is a {@link MapTreeSequenceField}.
 */
function isMapTreeSequenceField(field) {
    return field instanceof EagerMapTreeSequenceField;
}
exports.isMapTreeSequenceField = isMapTreeSequenceField;
/**
 * A readonly implementation of {@link FlexTreeNode} which wraps a {@link MapTree}.
 * @remarks Any methods that would mutate the node will fail,
 * as will the querying of data specific to the {@link LazyTreeNode} implementation (e.g. {@link FlexTreeNode.context}).
 * MapTreeNodes are unconditionally cached -
 * when retrieved via {@link getOrCreateNode}, the same {@link MapTree} object will always produce the same `MapTreeNode` object.
 */
class EagerMapTreeNode {
    /**
     * Create a new MapTreeNode.
     * @param location - the parentage of this node, if it is being created underneath an existing node and field, or undefined if not
     * @remarks This class (and its subclasses) should not be directly constructed outside of this module.
     * Instead, use {@link getOrCreateNode} to create a MapTreeNode from a {@link MapTree}.
     * A `MapTreeNode` may never be constructed more than once for the same {@link MapTree} object.
     * Instead, it should always be acquired via {@link getOrCreateNode}.
     */
    constructor(schema, 
    /** The underlying {@link MapTree} that this `MapTreeNode` reads its data from */
    mapTree, location = unparentedLocation) {
        this.schema = schema;
        this.mapTree = mapTree;
        this.location = location;
        this[_a] = index_js_3.FlexTreeEntityKind.Node;
        this.context = undefined;
        (0, internal_1.assert)(!nodeCache.has(mapTree), 0x98b /* A node already exists for the given MapTree */);
        nodeCache.set(mapTree, this);
        // Fully demand the tree to ensure that parent pointers are present and accurate on all nodes.
        // When a MapTreeNode is constructed, its MapTree may contain nodes (anywhere below) that map (via the `nodeCache`) to pre-existing MapTreeNodes.
        // Put another way, for a given MapTree, some ancestor MapTreeNode can be created after any number of its descendant MapTreeNodes already exist.
        // In such a case, the spine of nodes between the descendant and ancestor need to exist in order for the ancestor to be able to walk upwards via the `parentField` property.
        // This needs to happen for all MapTreeNodes that are descendants of the ancestor MapTreeNode.
        // Demanding the entire tree is overkill to solve this problem since not all descendant MapTree nodes will have corresponding MapTreeNodes.
        // However, demanding the full tree also lets us eagerly validate that there are no duplicate MapTrees (i.e. same MapTree object) anywhere in the tree.
        this.walkTree();
    }
    get type() {
        return this.schema.name;
    }
    adoptBy(parent, index) {
        if (parent !== undefined) {
            (0, internal_1.assert)(this.location === unparentedLocation, 0x98c /* Node may not be adopted if it already has a parent */);
            (0, internal_1.assert)(index !== undefined, 0xa08 /* Expected index */);
            this.location = { parent, index };
        }
        else {
            (0, internal_1.assert)(this.location !== unparentedLocation, 0xa09 /* Node may not be un-adopted if it does not have a parent */);
            this.location = unparentedLocation;
        }
    }
    /**
     * The field this tree is in, and the index within that field.
     * @remarks If this node is unparented, this method will return the special {@link unparentedLocation} as the parent.
     */
    get parentField() {
        return this.location;
    }
    is(schema) {
        return schema === this.schema;
    }
    tryGetField(key) {
        const field = this.mapTree.fields.get(key);
        // Only return the field if it is not empty, in order to fulfill the contract of `tryGetField`.
        if (field !== undefined && field.length > 0) {
            return getOrCreateField(this, key, this.schema.getFieldSchema(key));
        }
    }
    getBoxed(key) {
        const fieldKey = (0, index_js_2.brand)(key);
        return getOrCreateField(this, fieldKey, this.schema.getFieldSchema(fieldKey));
    }
    boxedIterator() {
        return (0, index_js_2.mapIterable)(this.mapTree.fields.entries(), ([key]) => getOrCreateField(this, key, this.schema.getFieldSchema(key)));
    }
    get value() {
        return this.mapTree.value;
    }
    get anchorNode() {
        // This API is relevant to `LazyTreeNode`s, but not `MapTreeNode`s.
        // TODO: Refactor the FlexTreeNode interface so that stubbing this out isn't necessary.
        return (0, index_js_2.fail)("MapTreeNode does not implement anchorNode");
    }
    walkTree() {
        for (const [key, mapTrees] of this.mapTree.fields) {
            const field = getOrCreateField(this, key, this.schema.getFieldSchema(key));
            for (let index = 0; index < field.length; index++) {
                const child = getOrCreateChild(mapTrees[index] ?? (0, internal_1.oob)(), this.schema.getFieldSchema(key).allowedTypes, { parent: field, index });
                // These next asserts detect the case where `getOrCreateChild` gets a cache hit of a different node than the one we're trying to create
                (0, internal_1.assert)(child.location !== undefined, 0x98d /* Expected node to have parent */);
                (0, internal_1.assert)(child.location.parent.parent === this, 0x98e /* Node may not be multi-parented */);
                (0, internal_1.assert)(child.location.index === index, 0x98f /* Node may not be multi-parented */);
                child.walkTree();
            }
        }
    }
}
exports.EagerMapTreeNode = EagerMapTreeNode;
_a = index_js_3.flexTreeMarker;
/**
 * The implementation of a map node created by {@link getOrCreateNode}.
 */
class EagerMapTreeMapNode extends EagerMapTreeNode {
    keys() {
        return this.mapTree.fields.keys();
    }
    values() {
        return (0, index_js_2.mapIterable)(this.mapTree.fields.keys(), (key) => unboxedField(this.tryGetField(key) ?? (0, index_js_2.fail)("Unexpected empty map field"), key, this.mapTree, this));
    }
    entries() {
        return (0, index_js_2.mapIterable)(this.mapTree.fields.keys(), (key) => [
            key,
            unboxedField(this.tryGetField(key) ?? (0, index_js_2.fail)("Unexpected empty map field"), key, this.mapTree, this),
        ]);
    }
    forEach(callbackFn, thisArg) {
        const fn = thisArg !== undefined ? callbackFn.bind(thisArg) : callbackFn;
        for (const [key, value] of this.entries()) {
            fn(value, key, this);
        }
    }
    getBoxed(key) {
        return super.getBoxed(key);
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    boxedIterator() {
        return super.boxedIterator();
    }
}
exports.EagerMapTreeMapNode = EagerMapTreeMapNode;
class EagerMapTreeLeafNode extends EagerMapTreeNode {
    get value() {
        return super.value;
    }
}
/**
 * A special singleton that is the implicit {@link LocationInField} of all un-parented {@link EagerMapTreeNode}s.
 * @remarks This exists because {@link EagerMapTreeNode.parentField} must return a field.
 * If a {@link EagerMapTreeNode} is created without a parent, its {@link EagerMapTreeNode.parentField} property will point to this object.
 * However, this field cannot be used in any practical way because it is empty, i.e. it does not actually contain the children that claim to be parented under it.
 * It has the "empty" schema and it will always contain zero children if queried.
 * Any nodes with this location will have a dummy parent index of `-1`.
 */
const unparentedLocation = {
    parent: {
        [index_js_3.flexTreeMarker]: index_js_3.FlexTreeEntityKind.Field,
        length: 0,
        key: index_js_1.EmptyKey,
        parent: undefined,
        is(schema) {
            return schema === index_js_4.FlexFieldSchema.empty;
        },
        boxedIterator() {
            return [].values();
        },
        boxedAt(index) {
            return undefined;
        },
        schema: index_js_4.FlexFieldSchema.empty,
        context: undefined,
        mapTrees: [],
    },
    index: -1,
};
class EagerMapTreeField {
    constructor(schema, key, parent) {
        this.schema = schema;
        this.key = key;
        this.parent = parent;
        this[_b] = index_js_3.FlexTreeEntityKind.Field;
        const fieldKeyCache = getFieldKeyCache(parent);
        (0, internal_1.assert)(!fieldKeyCache.has(key), 0x990 /* A field already exists for the given MapTrees */);
        fieldKeyCache.set(key, this);
        // When this field is created (which only happens one time, because it is cached), all the children become parented for the first time.
        // "Adopt" each child by updating its parent information to point to this field.
        for (const [i, mapTree] of this.mapTrees.entries()) {
            const mapTreeNodeChild = nodeCache.get(mapTree);
            if (mapTreeNodeChild !== undefined) {
                if (mapTreeNodeChild.parentField !== unparentedLocation) {
                    throw new internal_2.UsageError("A node may not be in more than one place in the tree");
                }
                mapTreeNodeChild.adoptBy(this, i);
            }
        }
    }
    get mapTrees() {
        return this.parent.mapTree.fields.get(this.key) ?? [];
    }
    get length() {
        return this.mapTrees.length;
    }
    is(schema) {
        return this.schema.equals(schema);
    }
    boxedIterator() {
        return this.mapTrees
            .map((m, index) => getOrCreateChild(m, this.schema.allowedTypes, {
            parent: this,
            index,
        }))
            .values();
    }
    boxedAt(index) {
        const i = (0, index_js_3.indexForAt)(index, this.length);
        if (i === undefined) {
            return undefined;
        }
        const m = this.mapTrees[i];
        if (m !== undefined) {
            return getOrCreateChild(m, this.schema.allowedTypes, {
                parent: this,
                index: i,
            });
        }
    }
    /**
     * Mutate this field.
     * @param edit - A function which receives the current `MapTree`s that comprise the contents of the field so that it may be mutated.
     * The function may mutate the array in place or return a new array.
     * If a new array is returned then it will be used as the new contents of the field, otherwise the original array will be continue to be used.
     * @remarks All edits to the field (i.e. mutations of the field's MapTrees) should be directed through this function.
     * This function ensures that the parent MapTree has no empty fields (which is an invariant of `MapTree`) after the mutation.
     */
    edit(edit) {
        const oldMapTrees = this.parent.mapTree.fields.get(this.key) ?? [];
        const newMapTrees = edit(oldMapTrees) ?? oldMapTrees;
        if (newMapTrees.length > 0) {
            this.parent.mapTree.fields.set(this.key, newMapTrees);
        }
        else {
            this.parent.mapTree.fields.delete(this.key);
        }
    }
}
_b = index_js_3.flexTreeMarker;
class EagerMapTreeOptionalField extends EagerMapTreeField {
    constructor() {
        super(...arguments);
        this.editor = {
            set: (newContent) => {
                // If the new content is a MapTreeNode, it needs to have its parent pointer updated
                if (newContent !== undefined) {
                    nodeCache.get(newContent)?.adoptBy(this, 0);
                }
                // If the old content is a MapTreeNode, it needs to have its parent pointer unset
                const oldContent = this.mapTrees[0];
                if (oldContent !== undefined) {
                    nodeCache.get(oldContent)?.adoptBy(undefined);
                }
                this.edit((mapTrees) => {
                    if (newContent !== undefined) {
                        mapTrees[0] = newContent;
                    }
                    else {
                        mapTrees.length = 0;
                    }
                });
            },
        };
    }
    get content() {
        const value = this.mapTrees[0];
        if (value !== undefined) {
            return unboxedUnion(this.schema, value, {
                parent: this,
                index: 0,
            });
        }
        return undefined;
    }
}
class EagerMapTreeRequiredField extends EagerMapTreeOptionalField {
    get content() {
        return super.content ?? (0, index_js_2.fail)("Expected EagerMapTree required field to have a value");
    }
}
class EagerMapTreeSequenceField extends EagerMapTreeField {
    constructor() {
        super(...arguments);
        this.editor = {
            insert: (index, newContent) => {
                for (let i = 0; i < newContent.length; i++) {
                    const c = newContent[i];
                    (0, internal_1.assert)(c !== undefined, 0xa0a /* Unexpected sparse array content */);
                    nodeCache.get(c)?.adoptBy(this, index + i);
                }
                this.edit((mapTrees) => {
                    if (newContent.length < 1000) {
                        // For "smallish arrays" (`1000` is not empirically derived), the `splice` function is appropriate...
                        mapTrees.splice(index, 0, ...newContent);
                    }
                    else {
                        // ...but we avoid using `splice` + spread for very large input arrays since there is a limit on how many elements can be spread (too many will overflow the stack).
                        return mapTrees.slice(0, index).concat(newContent, mapTrees.slice(index));
                    }
                });
            },
            remove: (index, count) => {
                for (let i = index; i < index + count; i++) {
                    const c = this.mapTrees[i];
                    (0, internal_1.assert)(c !== undefined, 0xa0b /* Unexpected sparse array */);
                    nodeCache.get(c)?.adoptBy(undefined);
                }
                let removed;
                this.edit((mapTrees) => {
                    removed = mapTrees.splice(index, count);
                });
                return removed ?? (0, index_js_2.fail)("Expected removed to be set by edit");
            },
        };
    }
    at(index) {
        const i = (0, index_js_3.indexForAt)(index, this.length);
        if (i === undefined) {
            return undefined;
        }
        return unboxedUnion(this.schema, this.mapTrees[i] ?? (0, internal_1.oob)(), { parent: this, index: i });
    }
    map(callbackfn) {
        return Array.from(this, callbackfn);
    }
    mapBoxed(callbackfn) {
        return Array.from(this.boxedIterator(), callbackfn);
    }
    *[Symbol.iterator]() {
        for (const [i, mapTree] of this.mapTrees.entries()) {
            yield unboxedUnion(this.schema, mapTree, { parent: this, index: i });
        }
    }
    getFieldPath() {
        throw unsupportedUsageError("Editing an array");
    }
}
// #endregion Fields
// #region Caching and unboxing utilities
const nodeCache = new WeakMap();
/** Node Parent -\> Field Key -\> Field */
const fieldCache = new WeakMap();
function getFieldKeyCache(parent) {
    return (0, index_js_2.getOrCreate)(fieldCache, parent, () => new Map());
}
/**
 * If there exists a {@link EagerMapTreeNode} for the given {@link MapTree}, returns it, otherwise returns `undefined`.
 * @remarks {@link EagerMapTreeNode | MapTreeNodes} are created via {@link getOrCreateNode}.
 */
function tryGetMapTreeNode(mapTree) {
    return nodeCache.get(mapTree);
}
exports.tryGetMapTreeNode = tryGetMapTreeNode;
/**
 * Create a {@link EagerMapTreeNode} that wraps the given {@link MapTree}, or get the node that already exists for that {@link MapTree} if there is one.
 * @param nodeSchema - the {@link FlexTreeNodeSchema | schema} that the node conforms to
 * @param mapTree - the {@link MapTree} containing the data for this node.
 * @remarks It must conform to the `nodeSchema`.
 */
function getOrCreateMapTreeNode(nodeSchema, mapTree) {
    return nodeCache.get(mapTree) ?? createNode(nodeSchema, mapTree, undefined);
}
exports.getOrCreateMapTreeNode = getOrCreateMapTreeNode;
/** Helper for creating a `MapTreeNode` given the parent field (e.g. when "walking down") */
function getOrCreateChild(mapTree, implicitAllowedTypes, parent) {
    const cached = nodeCache.get(mapTree);
    if (cached !== undefined) {
        return cached;
    }
    const allowedTypes = (0, schemaBuilderBase_js_1.normalizeAllowedTypes)(implicitAllowedTypes);
    const nodeSchema = allowedTypes
        .map((t) => ((0, index_js_4.isLazy)(t) ? t() : t))
        .find((t) => {
        (0, internal_1.assert)(t !== "Any", 0x993 /* 'Any' type is not supported */);
        return t.name === mapTree.type;
    }) ?? (0, index_js_2.fail)("Unsupported node schema");
    return createNode(nodeSchema, mapTree, parent);
}
/** Always constructs a new node, therefore may not be called twice for the same `MapTree`. */
function createNode(nodeSchema, mapTree, parentField) {
    if ((0, index_js_4.schemaIsLeaf)(nodeSchema)) {
        return new EagerMapTreeLeafNode(nodeSchema, mapTree, parentField);
    }
    if ((0, index_js_4.schemaIsMap)(nodeSchema)) {
        return new EagerMapTreeMapNode(nodeSchema, mapTree, parentField);
    }
    if ((0, index_js_4.schemaIsObjectNode)(nodeSchema)) {
        return new EagerMapTreeNode(nodeSchema, mapTree, parentField);
    }
    (0, internal_1.assert)(false, 0x994 /* Unrecognized node kind */);
}
/** Creates a field with the given attributes, or returns a cached field if there is one */
function getOrCreateField(parent, key, schema) {
    const cached = getFieldKeyCache(parent).get(key);
    if (cached !== undefined) {
        return cached;
    }
    if (schema.kind.identifier === index_js_5.FieldKinds.required.identifier ||
        schema.kind.identifier === index_js_5.FieldKinds.identifier.identifier) {
        return new EagerMapTreeRequiredField(schema, key, parent);
    }
    if (schema.kind.identifier === index_js_5.FieldKinds.optional.identifier) {
        return new EagerMapTreeOptionalField(schema, key, parent);
    }
    if (schema.kind.identifier === index_js_5.FieldKinds.sequence.identifier) {
        return new EagerMapTreeSequenceField(schema, key, parent);
    }
    return new EagerMapTreeField(schema, key, parent);
}
/** Unboxes non-polymorphic leaf nodes to their values, if applicable */
function unboxedUnion(schema, mapTree, parent) {
    const type = schema.monomorphicChildType;
    if (type !== undefined) {
        if ((0, index_js_4.schemaIsLeaf)(type)) {
            return mapTree.value;
        }
        return getOrCreateChild(mapTree, type, parent);
    }
    return getOrCreateChild(mapTree, schema.allowedTypes, parent);
}
/** Unboxes non-polymorphic required and optional fields holding leaf nodes to their values, if applicable */
function unboxedField(field, key, mapTree, parentNode) {
    const fieldSchema = field.schema;
    const mapTrees = mapTree.fields.get(key) ?? (0, index_js_2.fail)("Key does not exist in unhydrated map tree");
    if (fieldSchema.kind === index_js_5.FieldKinds.required) {
        return unboxedUnion(fieldSchema, mapTrees[0] ?? (0, internal_1.oob)(), {
            parent: field,
            index: 0,
        });
    }
    if (fieldSchema.kind === index_js_5.FieldKinds.optional) {
        return (mapTrees.length > 0
            ? unboxedUnion(fieldSchema, mapTrees[0] ?? (0, internal_1.oob)(), {
                parent: field,
                index: 0,
            })
            : undefined);
    }
    return getOrCreateField(parentNode, key, fieldSchema);
}
// #endregion Caching and unboxing utilities
function unsupportedUsageError(message) {
    return new internal_2.UsageError(`${message ?? "Operation"} is not supported for content that has not yet been inserted into the tree`);
}
exports.unsupportedUsageError = unsupportedUsageError;
//# sourceMappingURL=mapTreeNode.js.map