/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type Listenable, type Off } from "../../events/index.js";
import type { TreeChangeEvents, TreeNode, Unhydrated } from "./types.js";
import type { AnchorNode } from "../../core/index.js";
import { TreeStatus } from "../../feature-libraries/index.js";
import type { TreeNodeSchema } from "./treeNodeSchema.js";
export declare function getKernel(node: TreeNode): TreeNodeKernel;
/**
 * Detects if the given 'candidate' is a TreeNode.
 *
 * @remarks
 * Supports both Hydrated and {@link Unhydrated} TreeNodes, both of which return true.
 *
 * Because the common usage is to check if a value being inserted/set is a TreeNode,
 * this function permits calling with primitives as well as objects.
 *
 * Primitives will always return false (as they are copies of data, not references to nodes).
 *
 * @param candidate - Value which may be a TreeNode
 * @returns true if the given 'candidate' is a hydrated TreeNode.
 */
export declare function isTreeNode(candidate: unknown): candidate is TreeNode | Unhydrated<TreeNode>;
/**
 * Returns a schema for a value if the value is a {@link TreeNode}.
 *
 * Returns undefined for other values.
 * @remarks
 * Does not give schema for a {@link TreeLeafValue}.
 */
export declare function tryGetTreeNodeSchema(value: unknown): undefined | TreeNodeSchema;
/**
 * Contains state and an internal API for managing {@link TreeNode}s.
 * @remarks All {@link TreeNode}s have an associated kernel object.
 * The kernel has the same lifetime as the node and spans both its unhydrated and hydrated states.
 * When hydration occurs, the kernel is notified via the {@link TreeNodeKernel.hydrate | hydrate} method.
 */
export declare class TreeNodeKernel implements Listenable<TreeChangeEvents> {
    #private;
    readonly node: TreeNode;
    readonly schema: TreeNodeSchema;
    /**
     * Create a TreeNodeKernel which can be looked up with {@link getKernel}.
     * @remarks
     * Exactly one kernel per TreeNode should be created.
     */
    constructor(node: TreeNode, schema: TreeNodeSchema);
    hydrate(anchorNode: AnchorNode): void;
    dehydrate(): void;
    isHydrated(): boolean;
    getStatus(): TreeStatus;
    on<K extends keyof TreeChangeEvents>(eventName: K, listener: TreeChangeEvents[K]): Off;
    dispose(): void;
}
//# sourceMappingURL=treeNodeKernel.d.ts.map