"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryDisposeTreeNode = exports.setInnerNode = exports.tryGetCachedTreeNode = exports.tryGetInnerNode = exports.getOrCreateInnerNode = exports.anchorProxy = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../core/index.js");
const index_js_2 = require("../feature-libraries/index.js");
const index_js_3 = require("../util/index.js");
// TODO: decide how to deal with dependencies on flex-tree implementation.
// eslint-disable-next-line import/no-internal-modules
const lazyNode_js_1 = require("../feature-libraries/flex-tree/lazyNode.js");
const index_js_4 = require("./core/index.js");
// This file contains various maps and helpers for supporting associating simple TreeNodes with their InnerNodes, and swapping those InnerNodes as part of hydration.
// See ./ProxyBinding.md for a high-level overview of the process.
/**
 * An anchor slot which associates an anchor with its corresponding TreeNode, if there is one.
 * @remarks
 * For this to work, we have to require that there is at most a single view using a given AnchorSet.
 * FlexTree already has this assumption, and we also assume there is a single simple-tree per FlexTree, so this is valid.
 */
const proxySlot = (0, index_js_1.anchorSlot)();
// The following records are maintained as WeakMaps, rather than a private symbol (e.g. like `targetSymbol`) on the TreeNode.
// The map behaves essentially the same, except that performing a lookup in the map will not perform a property read/get on the key object (as is the case with a symbol).
// Since `SharedTreeNodes` are proxies with non-trivial `get` traps, this choice is meant to prevent the confusion of the lookup passing through multiple objects
// via the trap, or the trap not properly handling the special symbol, etc.
/**
 * A reverse mapping of {@link proxySlot} that is updated at the same time.
 *
 * @remarks
 * Nodes in this map are hydrated (and thus "marinated" or "cooked").
 * Nodes not in this map are known to be {@link Unhydrated}.
 * Thus this map can be used to check if a node is hydrated.
 *
 * Any node not in this map must be in {@link proxyToMapTreeNode} since it contains all unhydrated nodes.
 * It also contains "marinated" nodes which are in both.
 */
const proxyToAnchorNode = new WeakMap();
/**
 * Map {@link Unhydrated} nodes and "marinated" nodes to and from their underlying MapTreeNode.
 * These maps are populated by the TreeNode's constructor when called by a user before the node is inserted into the tree and queried.
 */
const proxyToMapTreeNode = new WeakMap();
/**
 * {@inheritdoc proxyToMapTreeNode}
 */
const mapTreeNodeToProxy = new WeakMap();
/**
 * Used by {@link anchorProxy} as an optimization to ensure that only one anchor is remembered at a time for a given anchor node
 */
const anchorForgetters = new WeakMap();
/**
 * Creates an anchor node and associates it with the given proxy.
 * @privateRemarks
 * Use `forgetters` to cleanup the anchor allocated by this function once the anchor is no longer needed.
 * In practice, this happens when either the anchor node is destroyed, or another anchor to the same node is created by a new flex node.
 *
 * The FlexTreeNode holds a reference to the same anchor, and has a lifetime at least as long as the simple-tree,
 * so this would be unnecessary except for the case of "marinated" nodes, which have an anchor,
 * but might not have a FlexTreeNode.
 * Handling this case is an optimization assuming that this extra anchor reference is cheaper than eagerly creating FlexTreeNodes.
 */
function anchorProxy(anchors, path, proxy) {
    (0, internal_1.assert)(!anchorForgetters.has(proxy), 0x91c /* Proxy anchor should not be set twice */);
    const anchor = anchors.track(path);
    const anchorNode = anchors.locate(anchor) ?? (0, index_js_3.fail)("Expected anchor node to be present");
    bindHydratedNodeToAnchor(proxy, anchorNode);
    const forget = () => {
        if (anchors.locate(anchor)) {
            anchors.forget(anchor);
        }
        anchorForgetters.delete(proxy);
        off();
    };
    anchorForgetters.set(proxy, forget);
    const off = anchorNode.on("afterDestroy", forget);
    return anchorNode;
}
exports.anchorProxy = anchorProxy;
function getOrCreateInnerNode(treeNode, allowFreed = false) {
    const anchorNode = proxyToAnchorNode.get(treeNode);
    if (anchorNode !== undefined) {
        // The proxy is bound to an anchor node, but it may or may not have an actual flex node yet
        const flexNode = anchorNode.slots.get(index_js_2.flexTreeSlot);
        if (flexNode !== undefined) {
            return flexNode; // If it does have a flex node, return it...
        } // ...otherwise, the flex node must be created
        const context = anchorNode.anchorSet.slots.get(index_js_2.ContextSlot) ?? (0, index_js_3.fail)("missing context");
        const cursor = context.checkout.forest.allocateCursor("getFlexNode");
        context.checkout.forest.moveCursorToPath(anchorNode, cursor);
        const newFlexNode = (0, lazyNode_js_1.makeTree)(context, cursor);
        cursor.free();
        // Calling this is a performance improvement, however, do this only after demand to avoid momentarily having no anchors to anchorNode
        anchorForgetters?.get(treeNode)?.();
        if (!allowFreed) {
            (0, index_js_2.assertFlexTreeEntityNotFreed)(newFlexNode);
        }
        return newFlexNode;
    }
    // Unhydrated case
    return proxyToMapTreeNode.get(treeNode) ?? (0, index_js_3.fail)("Expected raw tree node for proxy");
}
exports.getOrCreateInnerNode = getOrCreateInnerNode;
/**
 * Retrieves the InnerNode associated with the given target via {@link setInnerNode}, if any.
 * @remarks
 * If `target` is a unhydrated node, returns its MapTreeNode.
 * If `target` is a cooked node (or marinated but a FlexTreeNode exists) returns the FlexTreeNode.
 * If the target is not a node, or a marinated node with no FlexTreeNode for its anchor, returns undefined.
 */
function tryGetInnerNode(target) {
    // Calling 'WeakMap.get()' with primitives (numbers, strings, etc.) will return undefined.
    // This is in contrast to 'WeakMap.set()', which will throw a TypeError if given a non-object key.
    const anchorNode = proxyToAnchorNode.get(target);
    // Hydrated node case
    if (anchorNode !== undefined) {
        const flex = anchorNode.slots.get(index_js_2.flexTreeSlot);
        if (flex !== undefined) {
            // Cooked, or possible Marinated but something else cased the flex tree node to exist.
            return flex;
        }
        // Marinated case
        (0, internal_1.assert)(proxyToMapTreeNode.get(target) === undefined, 0xa05 /* marinated nodes should not have MapTreeNodes */);
        return undefined;
    }
    // Unhydrated node or not a node case:
    return proxyToMapTreeNode.get(target);
}
exports.tryGetInnerNode = tryGetInnerNode;
/**
 * Retrieves the proxy associated with the given flex node via {@link setInnerNode}, if any.
 */
function tryGetCachedTreeNode(flexNode) {
    if ((0, index_js_2.isMapTreeNode)(flexNode)) {
        // Unhydrated case
        return mapTreeNodeToProxy.get(flexNode);
    }
    // Hydrated case
    return flexNode.anchorNode.slots.get(proxySlot);
}
exports.tryGetCachedTreeNode = tryGetCachedTreeNode;
/**
 * Associate the given TreeNode and the given flex node.
 * @returns The node.
 * @remarks
 * This creates a 1:1 mapping between the tree node and InnerNode.
 * Either can be retrieved from the other via {@link getOrCreateInnerNode}/{@link tryGetInnerNode} or {@link tryGetCachedTreeNode}.
 * If the given proxy is already mapped to an flex node, the existing mapping will be overwritten.
 * If the given flex node is already mapped to a different proxy, this function will fail.
 */
function setInnerNode(node, innerNode) {
    const existingFlexNode = proxyToAnchorNode.get(node)?.slots.get(index_js_2.flexTreeSlot);
    (0, internal_1.assert)(existingFlexNode === undefined, 0x91d /* Cannot associate a flex node with multiple targets */);
    if ((0, index_js_2.isMapTreeNode)(innerNode)) {
        // Unhydrated case
        proxyToMapTreeNode.set(node, innerNode);
        mapTreeNodeToProxy.set(innerNode, node);
    }
    else {
        // Hydrated case
        (0, internal_1.assert)(tryGetCachedTreeNode(innerNode) === undefined, 0x7f5 /* Cannot associate an flex node with multiple targets */);
        bindHydratedNodeToAnchor(node, innerNode.anchorNode);
    }
    return node;
}
exports.setInnerNode = setInnerNode;
/**
 * Bi-directionally associates the given hydrated TreeNode to the given anchor node.
 * @remarks Cleans up mappings to {@link MapTreeNode} - it is assumed that they are no longer needed once the proxy has an anchor node.
 */
function bindHydratedNodeToAnchor(node, anchorNode) {
    // If the proxy currently has a raw node, forget it
    const mapTreeNode = proxyToMapTreeNode.get(node);
    if (mapTreeNode !== undefined) {
        proxyToMapTreeNode.delete(node);
        mapTreeNodeToProxy.delete(mapTreeNode);
    }
    // Once a proxy has been associated with an anchor node, it should never change to another anchor node
    (0, internal_1.assert)(!proxyToAnchorNode.has(node), 0x91e /* Proxy has already been bound to a different anchor node */);
    proxyToAnchorNode.set(node, anchorNode);
    // However, it's fine for an anchor node to rotate through different proxies when the content at that place in the tree is replaced.
    anchorNode.slots.set(proxySlot, node);
    (0, index_js_4.getKernel)(node).hydrate(anchorNode);
}
function tryDisposeTreeNode(anchorNode) {
    const treeNode = anchorNode.slots.get(proxySlot);
    if (treeNode !== undefined) {
        const kernel = (0, index_js_4.getKernel)(treeNode);
        kernel.dispose();
    }
}
exports.tryDisposeTreeNode = tryDisposeTreeNode;
//# sourceMappingURL=proxyBinding.js.map