/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type ITreeCursor, type ITreeCursorSynchronous } from "../../core/index.js";
import { TreeNodeSchemaBase } from "../../feature-libraries/index.js";
import { type JsonCompatible } from "../../util/index.js";
import type { IFluidHandle } from "@fluidframework/core-interfaces/internal";
/**
 * Used to read generic json compatible data as a tree in the JSON domain.
 * The returned tree will have a schema in the json domain as defined by {@link jsonRoot}.
 *
 * @returns an {@link ITreeCursorSynchronous} for a single {@link JsonCompatible}.
 */
export declare function singleJsonCursor(root: JsonCompatible): ITreeCursorSynchronous;
/**
 * Extract a JS object tree from the contents of the given ITreeCursor.
 * Assumes that ITreeCursor contains only unaugmented JsonTypes.
 */
export declare function cursorToJsonObject(reader: ITreeCursor): JsonCompatible;
/** Used to construct a {@link TypedJsonCursor} */
export interface TypedJsonCompatibleObject {
    [key: string]: TypedJsonCompatible;
    [typedJsonSymbol]: string | TreeNodeSchemaBase;
}
/** Used to construct a {@link TypedJsonCursor} */
export type TypedJsonCompatible = JsonCompatible | TypedJsonCompatibleObject | TypedJsonCompatible[] | IFluidHandle | JsonCompatible;
declare const typedJsonSymbol: unique symbol;
/**
 * A variant of {@link singleJsonCursor} which allows types to be provided for nodes.
 *
 * @remarks Types are optional, but if present will be used to derive the type of the node when the cursor is read.
 *
 * This cursor differs from singleJsonCursor in that it inlines arrays, (arrays are not boxed into an "array node" but are directly interpreted as sequence fields).
 *
 * It also allows Fluid handles as input.
 *
 * @example
 * ```ts
 * const cursor = typedJsonCursor({
 *   [typedJsonCursor.type]: Point,
 *   x: 3,
 *   y: 42
 * });
 * ```
 */
declare const singleTypedJsonCursor: {
    (root: TypedJsonCompatible): ITreeCursorSynchronous;
    type: typeof typedJsonSymbol;
};
export { singleTypedJsonCursor as typedJsonCursor };
//# sourceMappingURL=jsonCursor.d.ts.map