/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type AnchorNode, type AnchorSet, type UpPath } from "../core/index.js";
import { type FlexTreeNodeSchema, type FlexMapNodeSchema, type FlexObjectNodeSchema, type FlexTreeMapNode, type FlexTreeNode, type FlexTreeObjectNode, type FieldKinds, type FlexFieldSchema, type MapTreeNode } from "../feature-libraries/index.js";
import type { WithType, TreeNode } from "./core/index.js";
import type { TreeArrayNode } from "./arrayNode.js";
import type { TreeMapNode } from "./mapNode.js";
/**
 * Creates an anchor node and associates it with the given proxy.
 * @privateRemarks
 * Use `forgetters` to cleanup the anchor allocated by this function once the anchor is no longer needed.
 * In practice, this happens when either the anchor node is destroyed, or another anchor to the same node is created by a new flex node.
 *
 * The FlexTreeNode holds a reference to the same anchor, and has a lifetime at least as long as the simple-tree,
 * so this would be unnecessary except for the case of "marinated" nodes, which have an anchor,
 * but might not have a FlexTreeNode.
 * Handling this case is an optimization assuming that this extra anchor reference is cheaper than eagerly creating FlexTreeNodes.
 */
export declare function anchorProxy(anchors: AnchorSet, path: UpPath, proxy: TreeNode): AnchorNode;
/**
 * Retrieves the flex node associated with the given target via {@link setInnerNode}.
 * @remarks
 * For {@link Unhydrated} nodes, this returns the MapTreeNode.
 *
 * For hydrated nodes it returns a FlexTreeNode backed by the forest.
 * Note that for "marinated" nodes, this FlexTreeNode exists and returns it: it does not return the MapTreeNode which is the current InnerNode.
 */
export declare function getOrCreateInnerNode(treeNode: TypedNode<FlexObjectNodeSchema>, allowFreed?: true): InnerNode & FlexTreeObjectNode;
export declare function getOrCreateInnerNode(treeNode: TreeArrayNode, allowFreed?: true): InnerNode;
export declare function getOrCreateInnerNode(treeNode: TreeMapNode, allowFreed?: true): InnerNode & FlexTreeMapNode<FlexMapNodeSchema<string, FlexFieldSchema<typeof FieldKinds.optional>>>;
export declare function getOrCreateInnerNode(treeNode: TreeNode, allowFreed?: true): InnerNode;
/**
 * For "cooked" nodes this is a FlexTreeNode thats a projection of forest content.
 * For {@link Unhydrated} nodes this is a MapTreeNode.
 * For "marinated" nodes, some code (ex: getOrCreateInnerNode) returns the FlexTreeNode thats a projection of forest content, and some code (ex: tryGetInnerNode) returns undefined.
 *
 * @remarks
 * Currently MapTreeNode extends FlexTreeNode, and most code which can work with either just uses FlexTreeNode.
 * TODO: Code should be migrating toward using this type to distinguish to two use-cases.
 *
 * TODO: The inconsistent handling of "marinated" cases should be cleaned up.
 * Maybe getOrCreateInnerNode should cook marinated nodes so they have a proper InnerNode?
 */
export type InnerNode = FlexTreeNode | MapTreeNode;
/**
 * Retrieves the InnerNode associated with the given target via {@link setInnerNode}, if any.
 * @remarks
 * If `target` is a unhydrated node, returns its MapTreeNode.
 * If `target` is a cooked node (or marinated but a FlexTreeNode exists) returns the FlexTreeNode.
 * If the target is not a node, or a marinated node with no FlexTreeNode for its anchor, returns undefined.
 */
export declare function tryGetInnerNode(target: unknown): InnerNode | undefined;
/**
 * Retrieves the proxy associated with the given flex node via {@link setInnerNode}, if any.
 */
export declare function tryGetCachedTreeNode(flexNode: InnerNode): TreeNode | undefined;
/**
 * Associate the given TreeNode and the given flex node.
 * @returns The node.
 * @remarks
 * This creates a 1:1 mapping between the tree node and InnerNode.
 * Either can be retrieved from the other via {@link getOrCreateInnerNode}/{@link tryGetInnerNode} or {@link tryGetCachedTreeNode}.
 * If the given proxy is already mapped to an flex node, the existing mapping will be overwritten.
 * If the given flex node is already mapped to a different proxy, this function will fail.
 */
export declare function setInnerNode<TNode extends TreeNode>(node: TNode, innerNode: InnerNode): TNode;
/**
 * Given a node's schema, return the corresponding object in the proxy-based API.
 */
type TypedNode<TSchema extends FlexTreeNodeSchema> = TreeNode & WithType<TSchema["name"]>;
export declare function tryDisposeTreeNode(anchorNode: AnchorNode): void;
export {};
//# sourceMappingURL=proxyBinding.d.ts.map